""" Create Breakline:
Tool creates Z points along an input polyline.

Options include creating points at only vertices, at a specified chainage, or vertices and chainage

The tool also has the option of populating the Z attribute using a Raster layer pulled directly from the point
locations, or using a statistical analysis within a search radius.

Created by ES
"""

import os
import arcpy

# Must be run in foreground

# __________________________________________________________________________________________________
# Functions


def remove_duplicates(values):
    # Removes duplicate values from a list
    output = []
    seen = set()
    for value in values:
        if value not in seen:
            output.append(value)
            seen.add(value)
    return output


def create_vertice_chainages(polyline):
    # returns the chainages of the vertices along a polyline
    # polyline must be given in a search cursor format with 'SHAPE@'
    for vertices in polyline:    
        v_chainages = []
        for vertice in vertices:
            point = arcpy.Point(vertice.X, vertice.Y)
            v_chainage = polyline.measureOnLine(point)
            v_chainages.append(v_chainage)
    return v_chainages
    

def create_user_chainages(length, inc):
    # creates a chainage list using a defined length and increment
    chainage = 0
    chainages = []
    while chainage <= length:
        chainages.append(chainage)
        chainage += inc
    return chainages
    

def create_points(polyline_input, inc, method_type):
    # Creates geometry points along a polyline
    # polyline is a shape file
    # Increment is distance between user input points along polyline
    # method == TRUE: means only user chainages are used
    
    created_points = []
    with arcpy.da.SearchCursor(polyline_input, ['SHAPE@', 'OID@']) as rows:
        for row in rows:  # considers every polyline in layer individually
            arcpy.AddMessage("\nProcessing... feature FID {0}.".format(row[1]))
            length = row[0].getLength("PLANAR", "METERS")
            arcpy.AddMessage("Feature length: {0}.".format(length))
            
            # create points chainage list for polyline
            if method_type:  # create points at only use user defined chainages
                if not increment > 0:  # no increment or negative increment specified
                    arcpy.AddError("Must have a defined (positive) user chainage if ignoring polyline vertices")
                else:
                    user_chainages = create_user_chainages(length, inc)
                    user_chainages.append(length)  # make sure there's a point at the end of the line
                    chainages = remove_duplicates(user_chainages)
                    chainages = sorted(chainages)
                    arcpy.AddMessage("Created {0} points at user defined chainage...".format(len(user_chainages)))
            else:  # create point at vertices and user defined chainage
                vertice_chainages = create_vertice_chainages(row[0])
                arcpy.AddMessage("Created {0} points at polyline vertices...".format(len(vertice_chainages)))
                if increment > 0:  # adding user defined chainages
                    user_chainages = create_user_chainages(length, inc)
                    chainages = vertice_chainages + user_chainages
                    chainages = remove_duplicates(chainages)
                    chainages = sorted(chainages)
                    # gives number of user defined chainages after duplicates have been removed
                    arcpy.AddMessage("Created {0} points at user defined chainage..."
                                     .format(len(chainages) - len(vertice_chainages)))
                else:
                    chainages = vertice_chainages
            # create the points from chainage list along the polyline
            for chainage in chainages:
                created_point = row[0].positionAlongLine(chainage)
                created_points.append(created_point)
    return created_points
    

def create_vertice_list(polyline_input, inc):
    # Creates a list of vertices that can be used to create a polyline
    feature_list = []
    with arcpy.da.SearchCursor(polyline_input, ['SHAPE@', 'OID@']) as rows:  # consider each polyline in layer
        for row in rows:
            length = row[0].getLength("PLANAR", "METERS")
            
            # create chainage list
            if not inc > 0:  # only at vertices
                chainages = create_vertice_chainages(row[0])
            else:
                # at vertices and user defined. Even if 'method == True' still require vertices to retain polyline shape
                chainages = create_user_chainages(length, inc) + create_vertice_chainages(row[0])
                chainages = remove_duplicates(chainages)
                chainages = sorted(chainages)
            feature = []
            
            # use chainages to create vertice list but in terms of X, Y coords
            for chainage in chainages:
                vertice = [row[0].positionAlongLine(chainage).trueCentroid.X,
                           row[0].positionAlongLine(chainage).trueCentroid.Y]
                feature.append(vertice)
            
            feature_list.append(feature)
    return feature_list
                

def create_polyline(polyline_input, inc):
    # Creates a geometry Polyline
    # Polyline input is a shape file
    vertice_list = create_vertice_list(polyline_input, inc)
    polyline = []
    for feature in vertice_list:
        polyline.append(
            arcpy.Polyline(
                arcpy.Array([arcpy.Point(*coords) for coords in feature])))
    return polyline


def new_lyr_name(input_file):
    # Defines a new layer name for the output geometry point layer based on the input name
    file_name = os.path.splitext(os.path.basename(input_file))[0]
    if file_name[-2:] == "_P" or file_name[-2:] == "_L" or file_name[-2:] == "_R":
        new_lyr_path = (os.path.splitext(input_file)[0])[:-2] + "_P" + os.path.splitext(input_file)[1]
    else:
        new_lyr_path = (os.path.splitext(input_file)[0]) + "_P" + os.path.splitext(input_file)[1]
    return new_lyr_path
    

def get_elevation(input_points, input_raster, field):
    # This is used in place of Extract Multi Value if the user does not have the Spatial Analyst extension in Arc
    # inputPoints is a shapefile
    with arcpy.da.UpdateCursor(input_points, ['SHAPE@XY', 'OID@', field]) as rows:
        for row in rows:
            location = " ".join(map(str, row[0]))
            cell_value = arcpy.GetCellValue_management(input_raster, location, "1")
            cell_value = str(cell_value)
            if cell_value == 'NoData':
                row[2] = -99999
            else:
                row[2] = cell_value
            rows.updateRow(row)
    return input_points
    

def get_elevation_stat(input_points, input_raster, buffer, stat_type):
    # This will take points and get the statistics from a DEM within a search radius
    # inputPoints is a shape file
    # search radius is in metres
    # stat type must be 'MAXIMUM' 'MINIMUM' or 'MEAN'
    # must have spatial analyst extension

    # hopefully forces projection to be in planar units so search radius is always in metres or feet
    arcpy.env.outputCoordinateSystem = arcpy.SpatialReference(input_points[:-4] + ".prj")
    # perform statistical analysis on search radius
    arcpy.Buffer_analysis(input_points, os.path.join(os.path.dirname(input_points), '_buffer'), buffer)
    arcpy.sa.ZonalStatisticsAsTable(os.path.join(os.path.dirname(input_points), '_buffer.shp'),
                                    'FID', input_raster.dataSource,
                                    os.path.join(os.path.dirname(input_points), '_stats.dbf'),
                                    'DATA', statType.upper())

    # field calculator input - null data will be given a value of -99999
    if stat_type.upper() == 'MAXIMUM':
        expression = "getStat(!MAX!)"
    elif stat_type.upper() == 'MINIMUM':
        expression = "getStat(!MIN!)"
    elif stat_type.upper() == 'MEAN':
        expression = "getStat(!MEAN!)"
    codeblock = """def getStat(input):
        if input != None:
            output = input
        else:
            output = -99999
        return output"""

    # join fields then copy stat attribute to Z value
    # convert feature class (shp file) to feature layer for join operation
    feature_layer_points = arcpy.MakeFeatureLayer_management(input_points, "feature_layer")
    # convert table to table view for join operation
    table_view_points = arcpy.MakeTableView_management(os.path.join(os.path.dirname(input_points), '_stats.dbf'),
                                                       "table_view")
    arcpy.AddJoin_management(feature_layer_points, 'FID', table_view_points, 'FID_')
    arcpy.CalculateField_management(feature_layer_points, 'Z', expression, 'PYTHON_9.3', codeblock)
    arcpy.RemoveJoin_management(feature_layer_points)

    # tidy up excess files
    arcpy.Delete_management(os.path.join(os.path.dirname(input_points), '_buffer.shp'))
    arcpy.Delete_management(os.path.join(os.path.dirname(input_points), '_stats.dbf'))
    
    return input_points
# __________________________________________________________________________________________________
# Main code


if __name__ == "__main__":
    
    # Get attributes
    zline_input = arcpy.GetParameter(0)
    method = arcpy.GetParameter(1)  # method = TRUE: means only user chainages are used
    increment = arcpy.GetParameter(2)
    raster = arcpy.GetParameter(3)
    searchRadius = arcpy.GetParameter(4)
    statType = arcpy.GetParameter(5)
    
    # Define additional variables
    inputDataSource = zline_input.dataSource
    sr = arcpy.Describe(zline_input).spatialReference
    
    # Create Z points
    createdPoints = create_points(zline_input, increment, method)
    createPointsData = new_lyr_name(zline_input.dataSource)
    try:
        arcpy.CopyFeatures_management(createdPoints, createPointsData)
    except Exception:  # file name most likely has 'invalid characters'. Create file as temp then rename.
        arcpy.CopyFeatures_management(createdPoints, os.path.join(os.path.dirname(zline_input.dataSource), 'temp'))
        arcpy.Rename_management(os.path.join(os.path.dirname(zline_input.dataSource), 'temp.shp'), createPointsData)
    
    # Manage fields in points feature layer (TUFLOW Z shape format)
    arcpy.AddField_management(createPointsData, "Z", "DOUBLE", 15, 5, "#", "#", "#", "#", "#")
    arcpy.AddField_management(createPointsData, "dZ", "DOUBLE", 15, 5, "#", "#", "#", "#", "#")
    arcpy.AddField_management(createPointsData, "Shape_Widt", "DOUBLE", 15, 5, "#", "#", "#", "#", "#")
    arcpy.AddField_management(createPointsData, "Shape_Opti", "TEXT", "#", "#", 20, "#", "#", "#", "#")
    arcpy.DeleteField_management(createPointsData, "ID")
    
    # Extract elevation values if using a search radius
    if len(str(raster)) > 0 and searchRadius > 0:
        # input management, check stat type is acceptable
        if statType.upper() != 'MAXIMUM' and statType.upper() != 'MINIMUM' and statType.upper() != 'MEAN':
            if len(statType) == 0:
                statType = 'MEAN'  # default to mean if left blank
            else:
                arcpy.AddMessage("\nProcessing... Statistics type not recognised")
        try:
            get_elevation_stat(createPointsData, raster, searchRadius, statType)
            arcpy.AddMessage(
                "\nProcessing... Assigning elevation values using a {0}m search radius and calculating the {1} value"
                    .format(searchRadius, statType))
        except Exception:
            arcpy.AddWarning(
                "\nProcessing...Unexpected error when performing statistical search radius, check output. "
                "This error could be due to the Spatial Analyst extension- it is required to perform a statistical "
                "search radius analysis. Double check it is turned on.\n")
    # Extract values from raster DEM if not using statistics
    elif len(str(raster)) > 0:
        arcpy.AddMessage("\nProcessing... Assigning elevation values at point locations") 
        get_elevation(createPointsData, raster.dataSource, "Z")
    
    # Add created points layer to map
    mxd = arcpy.mapping.MapDocument("CURRENT")
    df = arcpy.mapping.ListDataFrames(mxd, "*")[0]
    newlayer = arcpy.mapping.Layer(createPointsData)
    arcpy.mapping.AddLayer(df, newlayer, "TOP")
    arcpy.RefreshActiveView()
    arcpy.RefreshTOC()
    del newlayer
    
    # Insert extra vertices along the polyline (replace old polyline with new polyline that has the required vertices)
    if increment > 0:
        # Create polyline
        createdPolylines = create_polyline(zline_input, increment)
        arcpy.CopyFeatures_management(createdPolylines,
                                      os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"))
        arcpy.DefineProjection_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"), sr)
        
        # Manage fields in polyline (TUFLOW Z shape format)
        arcpy.AddField_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"),
                                  "Z", "DOUBLE", 10, 5, "#", "#", "#", "#", "#")
        arcpy.AddField_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"),
                                  "dZ", "DOUBLE", 10, 5, "#", "#", "#", "#", "#")
        arcpy.AddField_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"),
                                  "Shape_Widt", "DOUBLE", 10, 5, "#", "#", "#", "#", "#")
        arcpy.AddField_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"),
                                  "Shape_Opti", "TEXT", "#", "#", 20, "#", "#", "#", "#")
        arcpy.DeleteField_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"), "ID")
                
        # Move input layer to superseded folder (just in case it's needed again) and replace with new polyline layer
        # To make the replace work, the old layer needs to be completely deleted before added again
        # (that's why it's a little messy)
        ssDir = os.path.join(zline_input.workspacePath, "SS")
        if not os.path.exists(ssDir):
            os.mkdir(ssDir)
        try:
            arcpy.CopyFeatures_management(zline_input, os.path.join(ssDir, os.path.basename(inputDataSource)))
        except Exception:  # file name most likely has 'invalid characters'. Create file as temp then rename.
            arcpy.CopyFeatures_management(zline_input, os.path.join(ssDir, 'temp'))
            arcpy.Rename_management(os.path.join(ssDir, 'temp.shp'), os.path.basename(inputDataSource))
        currentLyr = arcpy.mapping.ListLayers(mxd, zline_input.name)[0]
        currentLyr.replaceDataSource(zline_input.workspacePath, "SHAPEFILE_WORKSPACE", "_geomL")
        arcpy.Delete_management(inputDataSource)
        try:
            arcpy.CopyFeatures_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"),
                                          inputDataSource)
        except Exception:  # file name most likely has 'invalid characters'. Create file as temp then rename.
            arcpy.CopyFeatures_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"),
                                          os.path.join(os.path.dirname(zline_input.dataSource), 'temp'))
            arcpy.Rename_management(os.path.join(os.path.dirname(zline_input.dataSource), 'temp.shp'), inputDataSource)
        currentLyr = arcpy.mapping.ListLayers(mxd, zline_input.name)[0]
        currentLyr.replaceDataSource(zline_input.workspacePath, "SHAPEFILE_WORKSPACE",
                                     os.path.splitext(os.path.basename(inputDataSource))[0])
        arcpy.Delete_management(os.path.join(os.path.dirname(zline_input.dataSource), "_geomL.shp"))
        arcpy.RefreshTOC(), arcpy.RefreshActiveView()
    
    arcpy.AddMessage("\nSuccessfully created Z points\n")
