'''
Load Simulation Files from TLF File

version 2.0

author: Wood Rodgers, Inc.
    http://www.woodrodgers.com

Modified by ES 28/11/2017
- added ability to select which layer groups to add
'''

import os
import sys
import arcpy

# Get input parameters
in_tuflow = arcpy.GetParameterAsText(0)  # get location of *.tlf file
add_inputs = arcpy.GetParameter(1)  # boolean input: True = add input files
add_checks = arcpy.GetParameter(2)  # boolean input: True = add check files
add_results = arcpy.GetParameter(3)  # boolean input: True = add result files

# define paths to relevant folder locations
scriptPath = sys.path[0]
lyrPath = os.path.join(scriptPath.split('Scripts')[0], 'ToolData', 'LayerFiles')
input_lyrPath = os.path.join(lyrPath, 'Inputs')
check_lyrPath = os.path.join(lyrPath, 'Check')
result_lyrPath = os.path.join(lyrPath, 'TS_Res')

# define the correct Arc grouping layer
if add_inputs and add_checks and add_results:
    tuflow_GroupLayer = os.path.join(lyrPath, 'TUFLOW_ICR_(EmptyGroup).lyr')
elif add_inputs and add_checks:
    tuflow_GroupLayer = os.path.join(lyrPath, 'TUFLOW_IC_(EmptyGroup).lyr')
elif add_inputs and add_results:
    tuflow_GroupLayer = os.path.join(lyrPath, 'TUFLOW_IR_(EmptyGroup).lyr')
elif add_checks and add_results:
    tuflow_GroupLayer = os.path.join(lyrPath, 'TUFLOW_CR_(EmptyGroup).lyr')
elif add_inputs:
    tuflow_GroupLayer = os.path.join(lyrPath, 'TUFLOW_I_(EmptyGroup).lyr')
elif add_checks:
    tuflow_GroupLayer = os.path.join(lyrPath, 'TUFLOW_C_(EmptyGroup).lyr')
elif add_results:
    tuflow_GroupLayer = os.path.join(lyrPath, 'TUFLOW_R_(EmptyGroup).lyr')
else:
    tuflow_GroupLayer = os.path.join(lyrPath, 'TUFLOW_I_(EmptyGroup).lyr')


if __name__ == '__main__':
    # check there is something to add to the map
    if not add_inputs and not add_checks and not add_results:
        arcpy.AddError("\nProcessing...\nNo layers have been selected to add to the Map Window.\nExiting tool.\n")
    
    # set up mxd
    map_doc = "CURRENT"
    arcpy.env.workspace = os.path.dirname(map_doc)
    mxd = arcpy.mapping.MapDocument(map_doc)
    df = arcpy.mapping.ListDataFrames(mxd, "Layers")[0]
    
    # set up .lyr template with groups for adding shp files to
    groupLayer = arcpy.mapping.Layer(tuflow_GroupLayer)
    arcpy.mapping.AddLayer(df, groupLayer, "BOTTOM")
    if add_inputs:
        inputGroupLayer = arcpy.mapping.ListLayers(mxd, "INPUT LAYERS", df)[0]
    if add_checks:
        checkGroupLayer = arcpy.mapping.ListLayers(mxd, "CHECK LAYERS", df)[0]
    if add_results:
        resultGroupLayer = arcpy.mapping.ListLayers(mxd, "TS RESULT LAYERS", df)[0]
    
    # start section on reading in shp files from *.tlf
    arcpy.AddMessage('\nProcessing...\nReading TUFLOW *.tlf...')
    layer_list_1d = []
    layer_list_2d = []
    layer_list_check = []
    layer_list_result = []
    
    f = open(in_tuflow, "r")
    for line in f.readlines():
        if line.startswith('Opening GIS Layer:'):  # relevant lines in tlf specifying shp layers
            shp = line.split('Opening GIS Layer:')[1].strip()  # obtain full shp path
            
            # check file exists with current name
            if os.path.isfile(shp):
                # if exists, sort gis layer into relevant list
                if '\RESULTS' in shp.upper():
                    if shp not in layer_list_result:
                        layer_list_result.append(shp)
                elif '\CHECK' in shp.upper():
                    if shp not in layer_list_check: 
                        layer_list_check.append(shp)
                elif '1D' in shp.upper():
                    if shp not in layer_list_1d:
                        layer_list_1d.append(shp)
                elif '2D' in shp.upper():
                    if shp not in layer_list_2d:
                        layer_list_2d.append(shp)
                else:  # if unkown put in 2D layer list
                    if shp not in layer_list_2d:
                        layer_list_2d.append(shp)
            
            # else file may exist but requires additional geom code added ('_R' '_L' '_P')
            elif '\CHECK' in shp.upper():
                f1, f2 = os.path.splitext(shp)
                shp_R = f1 + '_R' + f2
                shp_L = f1 + '_L' + f2
                shp_P = f1 + '_P' + f2
                if os.path.isfile(shp_R):
                    if shp_R not in layer_list_check:
                        layer_list_check.append(shp_R)
                if os.path.isfile(shp_L):
                    if shp_L not in layer_list_check:
                        layer_list_check.append(shp_L)
                if os.path.isfile(shp_P):
                    if shp_P not in layer_list_check:
                        layer_list_check.append(shp_P)
            
            elif 'MESSAGES' in shp.upper():
                f1, f2 = os.path.splitext(shp)
                shp_P = f1 + '_P' + f2
                if os.path.isfile(shp_P):
                    if shp_P not in layer_list_check:
                        layer_list_check.append(shp_P)
                else:
                    arcpy.AddMessage('cannot find {0}'.format(shp_P))
            
            elif '\RESULTS' in shp.upper():
                f1, f2 = os.path.splitext(shp)
                shp_R = f1 + '_R' + f2
                shp_L = f1 + '_L' + f2
                shp_P = f1 + '_P' + f2
                if os.path.isfile(shp_R):
                    if shp_R not in layer_list_result:
                        layer_list_result.append(shp_R)
                if os.path.isfile(shp_L):
                    if shp_L not in layer_list_result:
                        layer_list_result.append(shp_L)
                if os.path.isfile(shp_P):
                    if shp_P not in layer_list_result:
                        layer_list_result.append(shp_P)
            
            else:  # still cannot find file
                arcpy.AddMessage('cannot find {0}'.format(shp))
    f.close()
    
    # clean up lists so that only shp files with data in them get read in
    for layer in layer_list_check:
        if arcpy.management.GetCount(layer)[0] == "0":
            layer_list_check.remove(layer)
            
    for layer in layer_list_result:
        if arcpy.management.GetCount(layer)[0] == "0":
            layer_list_result.remove(layer)
    arcpy.AddMessage('Finished reading *.tlf')
    
    # Add shp files to map layer
    arcpy.AddMessage('\nProcessing...\nAdding shp files to map...')
    if add_checks:
        for layer in sorted(layer_list_check):
            addLayer = arcpy.mapping.Layer(layer)
            arcpy.mapping.AddLayerToGroup(df, checkGroupLayer, addLayer, "BOTTOM")
    if add_inputs:
        for layer in sorted(layer_list_1d):
            addLayer = arcpy.mapping.Layer(layer)
            arcpy.mapping.AddLayerToGroup(df, inputGroupLayer, addLayer, "BOTTOM")
        for layer in sorted(layer_list_2d):
            addLayer = arcpy.mapping.Layer(layer)
            arcpy.mapping.AddLayerToGroup(df, inputGroupLayer, addLayer, "BOTTOM")
    if add_results:
        for layer in sorted(layer_list_result):
            addLayer = arcpy.mapping.Layer(layer)
            arcpy.mapping.AddLayerToGroup(df, resultGroupLayer, addLayer, "BOTTOM")
    
    # make sure zpt and uvpt layers aren't checked in TOC
    toc_layers = arcpy.mapping.ListLayers(mxd, "", df)
    for layer in toc_layers:
        if 'zpt' in layer.name or 'uvpt' in layer.name or 'grd' in layer.name:
            layer.visible = False
    arcpy.RefreshTOC()
    arcpy.RefreshActiveView()
    arcpy.AddMessage('Finished adding shp files to map')
    
    # apply symbology to layers in mxd
    arcpy.AddMessage('\nProcessing...\nApplying TUFLOW symbology to layers...')
    if add_inputs:
        input_styles = os.listdir(input_lyrPath)
        for lyr in arcpy.mapping.ListLayers(mxd, "INPUT LAYERS", df):
            if lyr.isGroupLayer:
                for slyr in lyr:
                    if str(slyr).split("_")[-1].upper() == 'P' or str(slyr).split("_")[-1].upper() == 'L' or \
                       str(slyr).split("_")[-1].upper() == 'R':
                        lyrFile = "_{0}_{1}_empty_{2}.lyr".format(str(slyr).split("_")[0].lower(),
                                                                str(slyr).split("_")[1].lower(),
                                                                str(slyr).split("_")[-1].upper())
                    else:
                        if arcpy.Describe(slyr).shapeType == 'Point':
                            geomType = 'P'
                        elif arcpy.Describe(slyr).shapeType == 'Polyline':
                            geomType = 'L'
                        elif arcpy.Describe(slyr).shapeType == 'Polygon':
                            geomType = 'R'
                        lyrFile = "_{0}_{1}_empty_{2}.lyr".format(str(slyr).split("_")[0].lower(),
                                                    str(slyr).split("_")[1].lower(),
                                                    geomType)
                    if lyrFile in input_styles:
                        updateLayer = arcpy.mapping.ListLayers(mxd, slyr.datasetName, df)[0]
                        sourceLayer = arcpy.mapping.Layer(os.path.join(input_lyrPath, lyrFile))
                        arcpy.mapping.UpdateLayer(df, updateLayer, sourceLayer, True)
                        if slyr.symbologyType == 'UNIQUE_VALUES':
                            if lyrFile == "_1d_nwk_empty_L.lyr" or lyrFile == "_1d_nwk_empty_P.lyr" or \
                               lyrFile == "_1d_mh_empty_P.lyr" or lyrFile == "_1d_nd_empty_P.lyr" or \
                               lyrFile == "_1d_nwkb_empty_L.lyr" or lyrFile == "_1d_nwke_empty_L.lyr" or \
                               lyrFile == "_1d_nwke_empty_P.lyr" or lyrFile == "_1d_pit_empty_P.lyr":
                                fields = arcpy.ListFields(slyr)
                                for i, field in enumerate(fields):
                                    if i == 3:
                                        value_field = field.name
                                slyr.symbology.valueField = value_field
                                slyr.symbology.addAllValues()
                            elif lyrFile == "_2d_bc_empty_L.lyr" or lyrFile == "_2d_bc_empty_P.lyr" or \
                                 lyrFile == "_2d_mat_empty_R.lyr" or lyrFile == "_2d_code_empty_R.lyr" or \
                                 lyrFile == "_2d_soil_empty_R.lyr" or lyrFile == '_1d_bc_empty_P.lyr' or \
                                 lyrFile == '_1d_bc_empty_R.lyr':
                                fields = arcpy.ListFields(slyr)
                                for i, field in enumerate(fields):
                                    if i == 2:
                                        value_field = field.name
                                slyr.symbology.valueField = value_field
                                slyr.symbology.addAllValues()
                            arcpy.RefreshTOC()
    
    if add_checks:
        check_styles = os.listdir(check_lyrPath)
        for lyr in arcpy.mapping.ListLayers(mxd, "CHECK LAYERS", df):
            if lyr.isGroupLayer:
                for slyr in lyr:
                    for check_style in check_styles:
                        if os.path.splitext(check_style)[0] in slyr.name:
                            updateLayer = arcpy.mapping.ListLayers(mxd, slyr.datasetName, df)[0]
                            sourceLayer = arcpy.mapping.Layer(os.path.join(check_lyrPath, check_style))
                            arcpy.mapping.UpdateLayer(df, updateLayer, sourceLayer, True)
    
    if add_results:
        result_styles = os.listdir(result_lyrPath)
        for lyr in arcpy.mapping.ListLayers(mxd, "TS RESULT LAYERS", df):
            if lyr.isGroupLayer:
                for slyr in lyr:
                    for result_style in result_styles:
                        if os.path.splitext(result_style)[0] in slyr.name:
                            updateLayer = arcpy.mapping.ListLayers(mxd, slyr.datasetName, df)[0]
                            sourceLayer = arcpy.mapping.Layer(os.path.join(result_lyrPath, result_style))
                            arcpy.mapping.UpdateLayer(df, updateLayer, sourceLayer, True)
    arcpy.AddMessage('Finished applying symbology\n')
